package bean.editor;

import bean.PropertyChangeMultiCaster;
import bean.annotation.Property;
import utils.Utils;

import javax.swing.table.AbstractTableModel;
import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * User: huojun
 * Date: 7-Sep-2006
 * Time: 1:35:10 PM
 */
public class BeanPropertyModel extends AbstractTableModel implements IBeanPropertyModel, ICategorizableTableModel {
    protected Object _bean;
    protected Class _class;
    protected boolean _readOnly;
    public static Class[] IgnoredPropertyTypes = new Class[] {Class.class};
    protected List<PropertyDescriptor> _pds = new ArrayList<PropertyDescriptor>();
    protected List<Class> _ignoredTypes = new ArrayList<Class>(Arrays.asList(IgnoredPropertyTypes));

    public BeanPropertyModel() {
        this(null);
    }

    public BeanPropertyModel(Object bean) {
        setBean(bean);
    }

    public Object getBean() {
        return _bean;
    }

    public List<Class> getIgnoredTypes() {
        return _ignoredTypes;
    }

    public boolean isReadOnly() {
        return _readOnly;
    }

    public void setReadOnly(boolean readOnly) {
        _readOnly = readOnly;
    }

    public void setIgnoredTypes(List<Class> ignoredTypes) {
        _ignoredTypes = ignoredTypes;
    }

    public void setBean(Object bean) {
        if (null != bean) {
            if (!bean.equals(_bean)) {
                _bean = bean;
                _class = bean.getClass();
                _pds.clear();
                try {
                    BeanInfo bi = Introspector.getBeanInfo(_class);
                    PropertyDescriptor[] pds = bi.getPropertyDescriptors();
                    for (PropertyDescriptor pd : pds) {
                        if (canExpose(pd.getReadMethod())
                                && canExpose(pd.getWriteMethod())
                                && !_ignoredTypes.contains(pd.getPropertyType())) {
                            _pds.add(pd);
                        }
                    }
                } catch (IntrospectionException e) {
//                    throw new RuntimeException(e);
                    Utils.debug("BeanPropertyModel.setBean: " + e);
                }
            }
        } else {
            _pds.clear();
        }
        fireTableStructureChanged();
    }

    private boolean canExpose(Method rm) {
        Property pa = getPropertyAnnotation(rm);
        return null == pa || pa.expose();
    }

    private Property getPropertyAnnotation(Method rm) {
        if (null != rm) {
            return rm.getAnnotation(Property.class);
        }
        return null;
    }

    public PropertyDescriptor getPropertyDescriptor(int row) {
        return _pds.get(row);
    }

    public int getRowCount() {
        return _pds.size();
    }

    public int getColumnCount() {
        return 2;
    }

    public String getColumnName(int columnIndex) {
        return columnIndex == 0 ? "Property" : "Value";
    }

    public Class<?> getColumnClass(int columnIndex) {
        return columnIndex == 0 ? PropertyDescriptor.class : super.getColumnClass(columnIndex);
    }

    public boolean isCellEditable(int rowIndex, int columnIndex) {
        Method wm = _pds.get(rowIndex).getWriteMethod();
        Method rm = _pds.get(rowIndex).getReadMethod();
        boolean editable = true;
        if( null != wm ) {
            Property a = wm.getAnnotation(Property.class);
            editable = a == null || a.editable();
        }
        return columnIndex == 1 && editable && ( (wm != null && !_readOnly)
                || rm.getReturnType().isArray() );
    }

    public Object getValueAt(int rowIndex, int columnIndex) {
        PropertyDescriptor pd = _pds.get(rowIndex);
        if (columnIndex == 1) {
            Method readMethod = pd.getReadMethod();
            try {
                return null == readMethod ? null : readMethod.invoke(_bean);
            } catch (Exception e) {
//                throw new RuntimeException(e);
                Utils.debug("BeanPropertyModel.getValueAt: " + e);
            }
        } else {
            return pd;
        }
        return null;
    }

    public void setValueAt(Object aValue, int rowIndex, int columnIndex) {
        PropertyDescriptor pd = _pds.get(rowIndex);
        try {
            PropertyChangeMultiCaster.getInstance().set(getBean(), pd, aValue);
        } catch (Exception e) {
            //throw new RuntimeException(e);
            Utils.debug("BeanPropertyModel.setValueAt: " + e);
        }
    }

    public String getCategory(int rowIndex) {
        PropertyDescriptor pd = _pds.get(rowIndex);
        Property pa = getPropertyAnnotation(pd.getReadMethod());
        if (pa == null || pa.category().length() == 0) {
            Property pa1 = getPropertyAnnotation(pd.getWriteMethod());
            if( null == pa1 || pa1.category().length() == 0 ) {
                return null;
            } else {
                return pa1.category();
            }
        } else {
            return pa.category();
        }
    }
}
