package dbhelp.db;

import bean.BeanUtils;
import utils.Utils;

import java.sql.*;
import java.util.ArrayList;
import java.util.List;
import java.util.Arrays;
import java.util.Collections;

/**
 * User: huojun
 * Date: 5-Sep-2006
 * Time: 1:56:15 PM
 */
public class DbMetaImpl implements IDbMeta {
    private Database _database;

    public DbMetaImpl(Database database) {
        setDatabase(database);
    }

    public Database getDatabase() {
        return _database;
    }

    public void setDatabase(Database database) {
        _database = database;
    }

    private Connection getInterConn() throws Exception {
        return ConnectionManager.getMetaConnection(_database);
    }

    public String getIdentifierQuoteString() throws Exception {
        try {
            if (getMeta() == null) {
                return null;
            } else {
                String qStr = getMeta().getIdentifierQuoteString();
                return " ".equals(qStr) ? "" : qStr;
            }
        } catch (Exception e) {
            _database.reportMetaError(e);
            throw e;
        }
    }

    public List<Catalog> getCatalogs() throws Exception {
        List<Catalog> list = getAllCatalogs();
        List<Catalog> items = new ArrayList<Catalog>();
        for(Catalog c : list) {
            if( !isFilteredCatalog(c.getName()) ) {
                items.add(c);
            }
        }
        return items;
    }

    public List<Catalog> getAllCatalogs() throws Exception {
        try {
            DatabaseMetaData meta = getMeta();
            if (null != meta) {
                ResultSet catalogs = meta.getCatalogs();
                String catalogTerm = meta.getCatalogTerm();
                List<Catalog> items = BeanUtils.rsToList(catalogs, Catalog.class);
                for (Catalog catalog : items) {
                    catalog.setVendorTerm(catalogTerm);
                    catalog.setDatabase(getDatabase());
                }
                return items;
            } else {
                return Collections.emptyList();
            }
        } catch (Exception e) {
            _database.reportMetaError(e);
            throw e;
        }
    }

    public List<Schema> getSchemas() throws Exception {
        try {
            DatabaseMetaData meta = getMeta();
            List<Schema> items = null;
            if (null != meta) {
                ResultSet schemas = meta.getSchemas();
                String schemaTerm = meta.getSchemaTerm();
                items = BeanUtils.rsToList(schemas, Schema.class);
                for (Schema schema : items) {
                    schema.setVendorTerm(schemaTerm);
                    schema.setDatabase(getDatabase());
                }
            }
            return items;
        } catch (Exception e) {
            _database.reportMetaError(e);
            throw e;
        }
    }

    public String getDatabaseProductName() throws Exception {
        try {
            return getMeta() == null ? null : getMeta().getDatabaseProductName();
        } catch (Exception e) {
            _database.reportMetaError(e);
            throw e;
        }
    }

    public String getDatabaseProductVersion() throws Exception {
        try {
            return getMeta() == null ? null : getMeta().getDatabaseProductVersion();
        } catch (Exception e) {
            _database.reportMetaError(e);
            throw e;
        }
    }

    public String getProcedureSyntax(String queryCmd) throws Exception {
        StringBuffer sb = new StringBuffer();
        Statement statement = null;
        try {
            statement = getInterConn().createStatement();
            ResultSet resultSet = statement.executeQuery(queryCmd);
            while (null != resultSet && resultSet.next()) {
                sb.append(resultSet.getString(1));
            }
        } finally {
            if (null != statement) {
                statement.close();
            }
        }
        return sb.toString();
    }

    public List<Index> getIndexes(Table table) throws Exception {
        try {
            DatabaseMetaData meta = getMeta();
            if (null != meta) {
                String tt = table.getTableType();
                if (!"SYNONYM".equals(tt)) {
                    ResultSet indexInfo = meta.getIndexInfo(table.getCatalog(), table.getSchema(), table.getName(), false, false);
                    return BeanUtils.rsToList(indexInfo, Index.class);
                }
            }
            return new ArrayList<Index>();
        } catch (Exception e) {
            _database.reportMetaError(e);
            throw e;
        }
    }

    public List<String> getTableTypes() throws Exception {
        ArrayList<String> list = new ArrayList<String>();
        try {
            DatabaseMetaData meta = getMeta();
            if (null != meta) {
                List<String> types = BeanUtils.rsToList(meta.getTableTypes(), String.class);
                for (String s : types) {
                    // ASE returns type with "  "
                    list.add(s.trim());
                }
                return list;
            }
            return list;
        } catch (Exception e) {
            _database.reportMetaError(e);
            throw e;
        }
    }

    private DatabaseMetaData getMeta() throws Exception {
        if (_database.isConnected()) {
            return getInterConn().getMetaData();
        } else {
            return null;
        }
    }

    public List<Table> getAllTables() throws Exception {
        try {
            return getTables(null, null, null, null);
        } catch (Exception e) {
            _database.reportMetaError(e);
            throw e;
        }
    }

    public List<Column> getColumns(Table t) throws Exception {
        try {
            DatabaseMetaData meta = getMeta();
            if (null != meta) {
                List<Column> columns = BeanUtils.rsToList(
                        meta.getColumns(t.getCatalog(), t.getSchema(), t.getName(), "%"), Column.class);
                t.setColumns(columns.toArray(new Column[columns.size()]));
                List<PrimaryKey> primaryKeys = getPrimaryKeys(t);
                for (PrimaryKey pk : primaryKeys) {
                    for (Column c : columns) {
                        if (pk.getColumnName().equals(c.getName())) {
                            c.setPrimaryKeyInfo(pk);
                        }
                    }
                }
                // Do not retrieve foreign keys for views 
                if (!"VIEW".equalsIgnoreCase(t.getTableType())) {
                    List<ForeignKey> foreignKeys = getForeignKeys(t);
                    for (ForeignKey fk : foreignKeys) {
                        for (Column c : columns) {
                            if (fk.getFkColumnName().equals(c.getName())) {
                                c.setForeignKeyInfo(fk);
                            }
                        }
                    }
                }
                for (Column column : columns) {
                    column.setTable(t);
                }
                return columns;
            } else {
                return new ArrayList<Column>();
            }
        } catch (Exception e) {
            _database.reportMetaError(e);
            throw e;
        }
    }

    public List<PrimaryKey> getPrimaryKeys(Table t) throws Exception {
        try {
            DatabaseMetaData meta = getMeta();
            if (null != meta) {
                ResultSet primaryKeys = meta.getPrimaryKeys(t.getCatalog(), t.getSchema(), t.getName());
                return BeanUtils.rsToList(primaryKeys, PrimaryKey.class);
            } else {
                return new ArrayList<PrimaryKey>();
            }
        } catch (Exception e) {
            _database.reportMetaError(e);
            throw e;
        }
    }

    public List<ForeignKey> getForeignKeys(Table t) throws Exception {
        try {
            DatabaseMetaData meta = getMeta();
            if (null != meta) {
                ResultSet importedKeys = meta.getImportedKeys(t.getCatalog(), t.getSchema(), t.getName());
                return BeanUtils.rsToList(importedKeys, ForeignKey.class);
            } else {
                return new ArrayList<ForeignKey>();
            }
        } catch (Exception e) {
            _database.reportMetaError(e);
            throw e;
        }
    }


    public List<ForeignKey> getExportedKeys(Table t) throws Exception {
        try {
            DatabaseMetaData meta = getMeta();
            if (null != meta) {
                ResultSet exportedKeys = meta.getExportedKeys(t.getCatalog(), t.getSchema(), t.getName());
                return BeanUtils.rsToList(exportedKeys, ForeignKey.class);
            } else {
                return new ArrayList<ForeignKey>();
            }
        } catch (Exception e) {
            _database.reportMetaError(e);
            throw e;
        }
    }

    public String[] getKeywords() throws Exception {
        try {
            DatabaseMetaData meta = getMeta();
            if (null != meta) {
                String sqlKeywords = meta.getSQLKeywords();
                return sqlKeywords == null ? new String[0] : sqlKeywords.split(",");
            } else {
                return new String[0];
            }
        } catch (SQLException e) {
            _database.reportMetaError(e);
            throw e;
        }
    }

    public List<Procedure> getProcedures(String catalog, String schema, String namePattern) throws Exception {
        if (null == catalog) {
            List<Catalog> catalogs = getCatalogs();
            if (catalogs.size() > 0) {
                List<Procedure> procedures = new ArrayList<Procedure>();
                for (Catalog c : catalogs) {
                    procedures.addAll(searchProcedures(c.getName(), schema, namePattern));
                }
                return procedures;
            }
        }
        return searchProcedures(catalog, schema, namePattern);
    }

    private List<Procedure> searchProcedures(String catalog, String schema, String namePattern) throws Exception {
        if (!isFilteredCatalog(catalog)) {
            try {
                DatabaseMetaData meta = getMeta();
                if (null != meta) {
                    ResultSet procedures = meta.getProcedures(catalog, schema, namePattern);
                    List<Procedure> procs = BeanUtils.rsToList(procedures, Procedure.class);
                    for (Procedure procedure : procs) {
                        procedure.setDatabase(_database);
                    }
                    return procs;
                }
            } catch (Exception e) {
                _database.reportMetaError(e);
                throw e;
            }
        }
        return Collections.emptyList();
    }

    private boolean isFilteredCatalog(String catalog) {
        String[] filters = _database.getCatalogFilter();
        return ( null != filters && filters.length > 0 ) && !Arrays.asList(filters).contains(catalog);
    }

    public List<ProcedureColumn> getProcedureColumns(Procedure p) throws Exception {
        try {
            DatabaseMetaData meta = getMeta();
            if (null != meta) {
                ResultSet procedureColumns = meta.getProcedureColumns(p.getCatalog(), p.getSchema(), p.getName(), "%");
                List<ProcedureColumn> columns = BeanUtils.rsToList(procedureColumns, ProcedureColumn.class);
                ProcedureColumn[] all = columns.toArray(new ProcedureColumn[columns.size()]);
                p.setParameters(all);
                p.setResultset(all);
                return columns;
            } else {
                return new ArrayList<ProcedureColumn>();
            }
        } catch (Exception e) {
            _database.reportMetaError(e);
            throw e;
        }
    }

    public List<Table> getTables(String catalog, String schema, String tableType, String tablePattern) throws Exception {
        if (null == catalog) {
            List<Table> tables = new ArrayList<Table>();
            List<Catalog> catalogs = getCatalogs();
            if (catalogs.size() > 0) {
                for (Catalog c : catalogs) {
                    tables.addAll(searchTables(c.getName(), schema, tableType, tablePattern));
                }
                return tables;
            }
        }
        return searchTables(catalog, schema, tableType, tablePattern);
    }

    private List<Table> searchTables(String catalog, String schema, String tableType, String tablePattern) throws Exception {
        if (!isFilteredCatalog(catalog)) {
            try {
                DatabaseMetaData meta = getMeta();
                List<Table> tables = new ArrayList<Table>();
                if (null != meta) {
                    try {
                        tables = BeanUtils.rsToList(meta.getTables(catalog, schema, tablePattern, tableType == null ? null : new String[]{tableType}), Table.class);
                        for (Table table : tables) {
                            table.setDatabase(_database);
                        }
                    } catch (Exception e) {
                        Utils.debug("DbMetaImpl.searchTables: " + e);
                    }
                }
                return tables;
            } catch (Exception e) {
                _database.reportMetaError(e);
                throw e;
            }
        }
        return Collections.emptyList();
    }

    public List<Function> getFunctions() throws Exception {
        try {
            List<Function> funs = new ArrayList<Function>();
            DatabaseMetaData meta = getMeta();
            if (null != meta) {
                funs.addAll(Function.listFromString("TimeDate Function", meta.getTimeDateFunctions()));
                funs.addAll(Function.listFromString("String Function", meta.getStringFunctions()));
                funs.addAll(Function.listFromString("System Function", meta.getSystemFunctions()));
            }
            return funs;
        } catch (SQLException e) {
            _database.reportMetaError(e);
            throw e;
        }
    }

}
