package dbhelp.db;

import dbhelp.db.model.*;
import dbhelp.db.ui.SQLParamEditor;
import utils.Log;
import utils.Utils;

import javax.swing.*;
import java.math.BigDecimal;
import java.sql.*;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;

/**
 * Created by IntelliJ IDEA.
 * User: David Huo
 * Date: Sep 15, 2006
 * Time: 11:08:04 PM
 */
public class SQLCommander {
    private Database _database;
    private Connection _interConn;
    private boolean _busy;
    private boolean _discard;

    public static void runSQLCommand(Database database, String sql, Table table, ISQLResultListener listener) {
        SQLCommander commander = database.getCommander();
        List<Object> params = new ArrayList<Object>();
        if ( checkParameter(database, sql, params) ) {
            if (!commander.isBusy()) {
                Thread t = new Thread(new SQLRunner(database, sql, table, listener, params.toArray()));
                t.start();
            } else {
                String defaultItem = "Try again later";
                int ret = JOptionPane.showOptionDialog(Utils.getFrame(), "There is a SQL statement is still running!",
                        "Confirm", JOptionPane.CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE, null,
                        new String[]{defaultItem, "Discard it and run the new one!"}, defaultItem);
                if (ret == 1) {
                    database.spawnCommander();
                    Thread t = new Thread(new SQLRunner(database, sql, table, listener, params.toArray()));
                    t.start();
                }
            }
        }
    }

    private static boolean checkParameter(Database database, String sql, List<Object> params) {
        SQLParamEditor editor = SQLParamEditor.editParameters(database, sql);
        boolean shouldRun;
        if (editor.hasParam()) {
            editor.setVisible(true);
            shouldRun = editor.isConfirmed();
            if( shouldRun ) {
                SQLParam[] sqlParams = editor.getParams();
                for (int i = 0; null != sqlParams && i < sqlParams.length; i++) {
                    SQLParam sqlParam = sqlParams[i];
                    try {
                        params.add( sqlParam.getObject(database) );
                    } catch (Exception e) {
                        Utils.error(e);
                        shouldRun = false;
                    }
                }
            }
        } else {
            shouldRun = true;
        }
        return shouldRun;
    }

    public static Object sureType(Object value, int sqlType, Database db) throws ParseException {
        if (null != value && value.toString().trim().length() == 0) {
            if (Types.VARCHAR != sqlType && Types.CHAR != sqlType) {
                value = null;
            }
        }
        if (null != value) {
            String strValue = value.toString();
            if (Types.INTEGER == sqlType) {
                return Integer.parseInt(strValue);
            } else if (Types.BIGINT == sqlType) {
                return Long.parseLong(strValue);
            } else if (Types.DOUBLE == sqlType) {
                return Double.parseDouble(strValue);
            } else if (Types.FLOAT == sqlType || Types.REAL == sqlType) {
                return Float.parseFloat(strValue);
            } else if (Types.SMALLINT == sqlType) {
                return Short.parseShort(strValue);
            } else if (Types.NUMERIC == sqlType || Types.DECIMAL == sqlType) {
                return new BigDecimal(strValue);
            } else {
                if (Types.DATE == sqlType) {
                    java.util.Date date = db.getDateFormatter().parse(strValue);
                    return new Date(date.getTime());
                } else if (Types.TIME == sqlType) {
                    java.util.Date date = db.getTimeFormatter().parse(strValue);
                    return new Time(date.getTime());
                } else if (Types.TIMESTAMP == sqlType) {
                    java.util.Date date = db.getDateTimeFormatter().parse(strValue);
                    return new Timestamp(date.getTime());
                }
            }
        } else {
            value = new NullParam(sqlType);
        }
        return value;
    }

    private static class SQLRunner implements Runnable {
        private Database _database;
        private String _sql;
        private Table _table;
        private ISQLResultListener _listener;
        private Object[] _params;

        public SQLRunner(Database database, String sql, Table table, ISQLResultListener listener, Object[] params) {
            _database = database;
            _sql = sql;
            _table = table;
            _listener = listener;
            _params = params;
        }

        public void run() {
            try {
                SQLCommander commander = _database.getCommander();
                List<ISQLResult> rms = commander.execute(_sql, _params);
                if (!commander.isDiscard()) {
                    if (null != _listener) {
                        SQLResultHolder resultHolder = new SQLResultHolder(rms);
                        resultHolder.setSqlCmd(_sql);
                        resultHolder.setTable(_table);
                        resultHolder.setDatabase(_database);
                        _listener.notifyResult(resultHolder);
                    }
                } else {
                    Utils.debug("Discarded SQLCommander: " + commander);
                    commander.close();
                }
            } catch (Exception e) {
                Utils.error(e);
            }
        }
    }

    public SQLCommander(Database database) {
        _database = database;
    }

    public Database getDatabase() {
        return _database;
    }

    public void close() {
        if (null != _interConn) {
            try {
                _interConn.close();
            } catch (SQLException e) {
                // ignore
            }
            _interConn = null;
        }
    }

    public boolean isDiscard() {
        return _discard;
    }

    public void setDiscard(boolean discard) {
        _discard = discard;
    }

    public SQLCommander spawn() {
        return new SQLCommander(_database);
    }

    public boolean isBusy() {
        return _busy;
    }

    public Object executeScalar(String cmd, Object... params) throws Exception {
        markBusy();
        try {
            PreparedStatement stmt = prepareParams(cmd, params, true);
            ResultSet rs = stmt.executeQuery();
            Object ret = null;
            if (rs.next()) {
                ret = rs.getString(1);
            }
            rs.close();
            stmt.close();
            return ret;
        } finally {
            _busy = false;
        }
    }

    private void markBusy() {
        if (_busy) {
            throw new SQLCommanderBusyException();
        }
        _busy = true;
    }

    public void discard() {
        if (_interConn != null) {
            try {
                _interConn.close();
            } catch (SQLException e) {
                //
            }
            _interConn = null;
            _busy = false;
        }
    }

    public List<ISQLResult> execute(String cmd, Object... params) throws Exception {
        markBusy();
        List<ISQLResult> l = new ArrayList<ISQLResult>();
        PreparedStatement stmt = null;
        try {
            stmt = prepareParams(cmd, params, false);
            boolean autoCommit = _database.isAutoCommit();
            Connection connection = getConnection();
            Utils.debug("autoCommit = " + autoCommit);
            connection.setAutoCommit(autoCommit);
            long beginTime = System.currentTimeMillis();
            boolean result = stmt.execute();

            long time = System.currentTimeMillis() - beginTime;
            String unit = "millisecond(s)";
            if (time / 1000 > 50) {
                unit = "second(s)";
                time = time / 1000;
            }
            Log.message("Execution time: %1$d %2$s", time, unit);

            int updateCount = stmt.getUpdateCount();
            while (result || (updateCount != -1)) {
                if (result) {
                    ResultSet r = stmt.getResultSet();
                    l.add(createTableModel(r, cmd, params));
                } else {
                    // process update count
                    l.add(new SQLResult(updateCount));
                }
                result = stmt.getMoreResults();
                updateCount = stmt.getUpdateCount();
            }
            System.out.println("l = " + l);
            return l;
        } finally {
            if (null != stmt) {
                try {
                    stmt.close();
                } catch (SQLException e) {
                    // ignore
                }
            }
            _busy = false;
        }
    }

    private ResultModel createTableModel(ResultSet resultSet, String cmd, Object[] params) throws Exception {
        ResultModel m = new ResultModel();
        m.setStatement(cmd);
        m.setParams(params);
        m.setDatabase(_database);
        m.loadResultSet(resultSet);
        return m;
    }

    private PreparedStatement prepareParams(String cmd, Object[] params, boolean internalCall) throws Exception {
        if (!internalCall) {
            Log.message("SQL: %1$s", cmd);
            Utils.debug("SQLCommander.prepareParams: " + cmd);
            Utils.dumpArray(params);
        }
        PreparedStatement statement = getConnection().prepareStatement(cmd);
        for (int i = 0; null != params && i < params.length; i++) {
            Object param = params[i];
            if (null != param) {
                if (param instanceof NullParam) {
                    statement.setNull(i + 1, ((NullParam) param).getType());
                } else if (!(param instanceof String)) {
                    statement.setObject(i + 1, param);
                } else {
                    statement.setString(i + 1, param.toString());
                }
            } else {
                statement.setNull(i + 1, Types.VARCHAR);
            }
        }
        return statement;
    }

    private Connection getConnection() throws Exception {
        if (null == _interConn) {
            _interConn = _database.getConnection();
        }
        return _interConn;
    }

    public void commit() throws SQLException {
        if (null != _interConn) {
            _interConn.commit();
        }
    }

    public void rollback() throws SQLException {
        if (null != _interConn) {
            _interConn.rollback();
        }
    }
}
