package dbhelp.db.model;

import bean.annotation.Property;
import dbhelp.db.*;
import utils.Config;
import utils.Utils;

import java.util.*;


/**
 * User: huojun
 * Date: 7-Sep-2006
 * Time: 10:41:31 AM
 */
public abstract class AbstractDBObject implements IDBNode {
    protected IDBNode _parent;
    protected int _iconIndex = -1;
    protected List<IDBNode> _children;
    protected Object _userObject;

    protected AbstractDBObject() {
    }

    protected AbstractDBObject(Object userObject) {
        _userObject = userObject;
    }

    @Property(expose = false)
    public int getIconIndex() {
        return _iconIndex;
    }

    protected static void addCatalogChildren(IDbMeta m, List<IDBNode> list, IDBNode parent) throws Exception {
        Database database = m.getDatabase();
        List<String> types = m.getTableTypes();
        String[] tableTypes = types.toArray(new String[types.size()]);
        Arrays.sort(tableTypes);
        for (String type : tableTypes) {
            TableTypeFolder o = new TableTypeFolder(new DBFilterObject(type));
            o.setParent(parent);
            o.setDatabase(database);
            list.add(o);
        }
        ProceduresNode procedures = new ProceduresNode(new DBFilterObject("Procedures"));
        procedures.setParent(parent);
        procedures.setDatabase(database);
        list.add(procedures);
        IMetaProvider mp = database.getMetaProviderInstance();
        FunctionsNode functions = new FunctionsNode();
        functions.setParent(parent);
        functions.setDataObject(database);
        list.add(functions);
        if (mp != null) {
            list.addAll(mp.getNodes(parent));
        }
    }

    public static Object[] getNodePath(IDBNode n) {
        List<IDBNode> path = new ArrayList<IDBNode>();
        path.add(0, n);
        IDBNode parent = n.getParent();
        while (null != parent) {
            path.add(0, parent);
            parent = parent.getParent();
        }
        return path.toArray();
    }

    public void setIconIndex(int iconIndex) {
        _iconIndex = iconIndex;
    }

    public Object getUserObject() {
        return _userObject;
    }

    public IDBNode getParent() {
        return _parent;
    }

    public void setParent(IDBNode parent) {
        _parent = parent;
    }

    protected List<IDBNode> populateChildren() throws Exception {
        return null;
    }

    @Property(expose = false)
    public List<IDBNode> getChildren() {
        if (_children == null || _children.size() == 0) {
            try {
                _children = populateChildren();
            } catch (Throwable e1) {
                Utils.debug("AbstractDBObject.getChildren: " + e1);
            }
        }
        return _children;
    }

    public void refresh() {
        _children = null;
        Database db = getDatabase();
        if (null == db || db.isConnected()) {
            getChildren();
        }
    }

    public String toString() {
        return null == _userObject ? null : _userObject.toString();
    }


    protected IDbMeta getDatabaseMetaData() {
        Database db = getDatabase();
        return null == db ? null : db.getMetaData();
    }

    public Database getDatabase() {
        if (getUserObject() instanceof Database) {
            return ((Database) getUserObject());
        }
        IDBNode parent = getParent();
        while (null != parent) {
            if (parent.getUserObject() instanceof Database) {
                return (Database) parent.getUserObject();
            }
            parent = parent.getParent();
        }
        return null;
    }

    public static class CatalogNode extends AbstractDBObject {
        public CatalogNode(Catalog c) {
            super(c);
            setIconIndex(1);
        }

        protected List<IDBNode> populateChildren() throws Exception {
            getCatalog();
            List<IDBNode> list = new ArrayList<IDBNode>();
            addCatalogChildren(getDatabase().getMetaData(), list, this);
            return list;
        }

        private Catalog getCatalog() {
            return (Catalog) getUserObject();
        }


        public String toString() {
            Catalog c = getCatalog();
            String term = c.getVendorTerm();
            term = Utils.hasValue(term) ? term + ": " : "";
            return term + c.getName();
        }
    }

    public static class IndexNode extends AbstractDBObject {
        public IndexNode(Object userObject) {
            super(userObject);
            _iconIndex = 8;
        }
    }

    public static class IndexesNode extends NamedFolder implements ILeafControl {
        public IndexesNode(Table t) {
            super("Indexes");
            setDataObject(t);
        }

        protected List<IDBNode> populateChildren() throws Exception {
            Table t = (Table) getDataObject();
            List<Index> indexes = t.getDatabase().getMetaData().getIndexes(t);
            List<IDBNode> nodes = new ArrayList<IDBNode>();
            for (Index index : indexes) {
                IndexNode n = new IndexNode(index);
                n.setParent(this);
                nodes.add(n);
            }
            return nodes;
        }

        public boolean isLeaf() {
            return false;
        }
    }

    public static class ForeignKeyNode extends AbstractDBObject {
        public ForeignKeyNode(ForeignKey fk) {
            super(fk);
            _iconIndex = 4;
        }
    }

    public static class ForeignKeysNode extends NamedFolder implements ILeafControl {

        public ForeignKeysNode(Table t) {
            super("Foreign Keys");
            setDataObject(t);
        }

        protected List<IDBNode> populateChildren() throws Exception {
            List<IDBNode> list = new ArrayList<IDBNode>();
            Table t = (Table) getDataObject();
            for (ForeignKey fk : t.getForeignKeys()) {
                ForeignKeyNode o = new ForeignKeyNode(fk);
                list.add(o);
                o.setParent(this);
            }
            return list;
        }

        public boolean isLeaf() {
            return false;
        }
    }

    public static class ProfilesNode extends AbstractDBObject {
        public ProfilesNode() {
            super(new DatabasesObject());
            _iconIndex = 14;
        }

        public List<IDBNode> populateChildren() throws Exception {
            List<IDBNode> profiles = new ArrayList<IDBNode>();
            for (Database db : Config.getConfig().getDatabases()) {
                ProfileNode n = new ProfileNode(db);
                n.setParent(this);
                profiles.add(n);
            }
            return profiles;
        }
    }

    public static abstract class NamedFolder extends AbstractDBObject {
        protected Object _dataObject;

        public NamedFolder(String name) {
            super(name);
        }

        public Object getDataObject() {
            return _dataObject;
        }

        public void setDataObject(Object dataObject) {
            _dataObject = dataObject;
        }
    }

    public static class TableTypeFolder extends AbstractDBObject implements ILeafControl {
        private List<Table> _tables;
        protected IDBNodeComparator IDBNodeComparator = new IDBNodeComparator();

        public TableTypeFolder(DBFilterObject f) {
            super(f);
        }

        public void setDatabase(Database db) {
            DBFilterObject f = (DBFilterObject) getUserObject();
            f.setDatabase(db);
            f.setNode(this);
        }

        public List<Table> getTables() {
            if (null == _tables) {
                try {
                    populateChildren();
                } catch (Exception e) {
                    // ignore
                }
            }
            return _tables;
        }

        public String toString() {
            DBFilterObject f = (DBFilterObject) getUserObject();
            Object[] fos = f.getFilterObjects();
            return f.toString() + (fos == null || fos.length == 0 ? "" : ", filtered by: " + Utils.arrayToString(fos));
        }

        public List<IDBNode> populateChildren() throws Exception {
            List<IDBNode> nodes = new ArrayList<IDBNode>();
            IDbMeta md = getDatabaseMetaData();
            if (null != md) {
                String tableType = getUserObject().toString();
                Object uo = getParent().getUserObject();
                String catalog = uo instanceof Catalog ? ((Catalog) uo).getName() : null;
                _tables = md.getTables(catalog, null, tableType, null);
                DBFilterObject fn = (DBFilterObject) getUserObject();
                for (Table t : _tables) {
                    if (fn.accept(t.getCatalog(), t.getSchema())) {
                        TableNode o = new TableNode(t);
                        o.setParent(this);
                        nodes.add(o);
                    }
                }
            }
            Collections.sort(nodes, IDBNodeComparator);
            return nodes;
        }

        public boolean isLeaf() {
            return false;
        }

    }

    public static class ColumnFolder extends NamedFolder {
        public ColumnFolder() {
            super("Columns");
        }

        public List<IDBNode> populateChildren() throws Exception {
            Table t = (Table) getParent().getUserObject();
            getDatabaseMetaData().getColumns(t);
            List<IDBNode> list = new ArrayList<IDBNode>();
            for (Column c : t.getColumns()) {
                ColumnNode o = new ColumnNode(c);
                o.setParent(this);
                list.add(o);
            }
            return list;
        }
    }

    public static class ColumnNode extends AbstractDBObject {
        public ColumnNode(Column c) {
            super(c);
            _iconIndex = c.isPrimaryKey() && c.isForeignKey() ? 6
                    : (c.isPrimaryKey() ? 2 : (c.isForeignKey() ? 4 : 3));
        }
    }

    public static class ReferencesNode extends NamedFolder implements ILeafControl {
        public ReferencesNode(Table t) {
            super("References");
            setDataObject(t);
        }

        protected List<IDBNode> populateChildren() throws Exception {
            List<IDBNode> list = new ArrayList<IDBNode>();
            Table t = (Table) getDataObject();
            for (Table f : t.getReferencedTables()) {
                TableNode o = new TableNode(f);
                o.setParent(this);
                list.add(o);

            }
            return list;
        }

        public boolean isLeaf() {
            return false;
        }
    }

    public static class TableNode extends AbstractDBObject implements ILeafControl {
        public TableNode(Table t) {
            super(t);
            _iconIndex = 5;
        }

        public List<IDBNode> populateChildren() {
            ArrayList<IDBNode> list = new ArrayList<IDBNode>();
            ColumnFolder o = new ColumnFolder();
            o.setParent(this);
            list.add(o);
            Table t = getTable();
            String tt = t.getTableType();
            if (!("VIEW".equalsIgnoreCase(tt) || "ALIAS".equalsIgnoreCase(tt) || "SYNONYM".equalsIgnoreCase(tt))) {
                Index[] indexes = t.getIndexes();
                if (indexes != null && indexes.length > 0) {
                    IndexesNode indexesNode = new IndexesNode(t);
                    indexesNode.setParent(this);
                    list.add(indexesNode);
                }
                ForeignKey[] foreignKeys = t.getForeignKeys();
                if (null != foreignKeys && foreignKeys.length > 0) {
                    ForeignKeysNode fks = new ForeignKeysNode(t);
                    fks.setParent(this);
                    list.add(fks);
                }
                Table[] referencedTables = t.getReferencedTables();
                if (referencedTables != null && referencedTables.length > 0) {
                    ReferencesNode rn = new ReferencesNode(t);
                    rn.setParent(this);
                    list.add(rn);
                }
            }
            return list;
        }

        public Table getTable() {
            return (Table) getUserObject();
        }

        public boolean isLeaf() {
            return false;
        }
    }

    public static class ProfileNode extends AbstractDBObject {
        public ProfileNode(Database database) {
            super(database);
        }


        @Property(expose = false)
        public int getIconIndex() {
            Database db = (Database) getUserObject();
            String[] filter = db.getCatalogFilter();
            return filter == null || filter.length == 0 ? 32 : 33;
        }

        public List<IDBNode> populateChildren() throws Exception {
            List<IDBNode> list = new ArrayList<IDBNode>();
            Database database = (Database) getUserObject();
            if (database.isConnected()) {
                IDbMeta m = getDatabaseMetaData();
                List<Catalog> catalogs = m.getCatalogs();
                if (catalogs.size() > 0) {
                    for (Catalog c : catalogs) {
                        CatalogNode cn = new CatalogNode(c);
                        list.add(cn);
                        cn.setParent(this);
                    }
                } else {
                    addCatalogChildren(m, list, this);
                }
            }
            return list;
        }
    }

    public static class ProcedureParams extends NamedFolder {
        public ProcedureParams(ProcedureNode procedureNode) {
            super("Parameters");
            setParent(procedureNode);
        }

        public List<IDBNode> populateChildren() throws Exception {
            List<IDBNode> list = new ArrayList<IDBNode>();
            Procedure p = (Procedure) getDataObject();
            ProcedureColumn[] params = p.getParameters();
            for (ProcedureColumn procedureColumn : params) {
                list.add(new ProcedureColumnNode(procedureColumn));
            }
            return list;
        }
    }

    public static class ProcedureColumnNode extends AbstractDBObject {
        public ProcedureColumnNode(Object userObject) {
            super(userObject);
            _iconIndex = 3;
        }
    }

    public static class ProcedureResult extends NamedFolder {
        public ProcedureResult(ProcedureNode procedureNode) {
            super("Result");
            setParent(procedureNode);
        }

        public List<IDBNode> populateChildren() throws Exception {
            List<IDBNode> list = new ArrayList<IDBNode>();
            Procedure p = (Procedure) getDataObject();
            ProcedureColumn[] resultset = p.getResultset();
            for (ProcedureColumn procedureColumn : resultset) {
                list.add(new ProcedureColumnNode(procedureColumn));
            }
            return list;
        }
    }

    public static class ProcedureNode extends AbstractDBObject implements ILeafControl {
        public ProcedureNode(Object userObject) {
            super(userObject);
            _iconIndex = 0;
        }

        public List<IDBNode> populateChildren() throws Exception {
            List<IDBNode> list = new ArrayList<IDBNode>();
            Procedure p = (Procedure) getUserObject();
            ProcedureColumn[] parameters = p.getParameters();
            if (parameters != null && parameters.length > 0) {
                ProcedureParams o = new ProcedureParams(this);
                o.setDataObject(p);
                list.add(o);
            }
            ProcedureColumn[] rs = p.getResultset();
            if (rs != null && rs.length > 0) {
                ProcedureResult o = new ProcedureResult(this);
                o.setDataObject(p);
                list.add(o);
            }
            return list;
        }

        public boolean isLeaf() {
            return false;
        }
    }

    public static class ProceduresNode extends TableTypeFolder implements ILeafControl {
        public ProceduresNode(DBFilterObject f) {
            super(f);
        }

        public List<IDBNode> populateChildren() throws Exception {
            List<IDBNode> list = new ArrayList<IDBNode>();
            DBFilterObject f = (DBFilterObject) getUserObject();
            if (f.getDatabase().isConnected()) {
                Object uo = getParent().getUserObject();
                String catalog = uo instanceof Catalog ? ((Catalog) uo).getName() : null;
                for (Procedure p : getDatabaseMetaData().getProcedures(catalog, null, null)) {
                    if (f.accept(p.getCatalog(), p.getSchema())) {
                        ProcedureNode o = new ProcedureNode(p);
                        o.setParent(this);
                        list.add(o);
                    }
                }
            }
            return list;
        }

        public boolean isLeaf() {
            return false;
        }
    }

    public static class FunctionNode extends AbstractDBObject {
        public FunctionNode(Object userObject) {
            super(userObject);
            _iconIndex = 0;
        }
    }

    public static class FunctionsNode extends AbstractDBObject.NamedFolder {
        public FunctionsNode() {
            super("Reported Functions");
        }

        public List<IDBNode> populateChildren() throws Exception {
            List<IDBNode> list = new ArrayList<IDBNode>();
            if (((Database) getDataObject()).isConnected()) {
                for (Function p : getDatabaseMetaData().getFunctions()) {
                    FunctionNode o = new FunctionNode(p);
                    o.setParent(this);
                    list.add(o);
                }
            }
            return list;
        }
    }

    private static class IDBNodeComparator implements Comparator<IDBNode> {
        public int compare(IDBNode o1, IDBNode o2) {
            return o1.toString().compareTo(o2.toString());
        }
    }
}

