package dbhelp.db.model;

import dbhelp.db.*;
import utils.Log;
import utils.Utils;

import javax.swing.table.AbstractTableModel;
import java.sql.*;
import java.util.*;


/**
 * User: David Huo
 * Date: Sep 15, 2006
 * Time: 11:09:52 PM
 */
public class ResultModel extends AbstractTableModel implements ISQLResult {
    private Database _database;
    private Object[] _params;
    private List<Row> _rows = new ArrayList<Row>();
    private List<Row> _deletedRows = new ArrayList<Row>();
    private String[] _columnNames;
    private String _statement;
    private Column[] _columns;
    private Table _table = null;
    private Map<String, Integer> _nameMap = new HashMap<String, Integer>();
    private boolean _incomplete;
    private boolean _editable;
    private String _tableName = null;

    public ResultModel() {
    }

    public ResultModel(ResultSet r) throws Exception {
        loadResultSet(r);
    }

    public boolean isIncomplete() {
        return _incomplete;
    }

    public boolean isEditable() {
        return _editable;
    }

    public void setEditable(boolean editable) {
        _editable = editable;
    }

    public String getName() {
        String tableName = null == _table ? _tableName : _table.getName();
        String s = !Utils.hasValue(tableName) ? "Data" : tableName;
        return s + (_incomplete ? "(x)" : "");
    }

    public void loadResultSet(ResultSet r) throws Exception {
        ResultSetMetaData m = r.getMetaData();
        int cc = m.getColumnCount();
        _columnNames = new String[cc];
        _columns = new Column[cc];
        Set<String> tableNames = new HashSet<String>();
        for (int c = 0; c < cc; c++) {
            int columnId = c + 1;
            String columnName = m.getColumnName(columnId);
            _nameMap.put(columnName, c);
            _columnNames[c] = columnName;
            Column dc = new Column();
            dc.setName(columnName);
            try {
                String tableName = m.getTableName(columnId);
                tableNames.add(tableName == null ? "" : tableName);
                if (Utils.hasValue(tableName)) {
                    Table t = new Table(tableName);
                    dc.setTable(t);
                }
            } catch (Exception e) {
                tableNames.add("");
            }
            dc.setType(m.getColumnType(columnId));
            dc.setColumnSize(m.getColumnDisplaySize(columnId));
            try {
                dc.setAutoIncrement(m.isAutoIncrement(columnId));
                dc.setDecimalDigits(m.getPrecision(columnId));
            } catch (Exception e) {
                // oracle bug: getPrecision cause exception
                Utils.debug("ResultModel.loadResultSet: " + e);
            }
            _columns[c] = dc;
        }
        _tableName = tableNames.size() == 1 && !tableNames.contains("") ? tableNames.iterator().next() : null;
        int rows = 0;
        while (r.next()) {
            Row row = new Row(cc);
            for (int i = 0; i < cc; i++) {
                Column c = _columns[i];
                int ct = c.getType();
                String strValue;
                if (ct == Types.DATE) {
                    java.sql.Date date = r.getDate(i + 1);
                    strValue = r.wasNull() ? null : getDatabase().getDateFormatter().format(date);
                } else if (ct == Types.TIME) {
                    Time time = r.getTime(i + 1);
                    strValue = r.wasNull() ? null : getDatabase().getTimeFormatter().format(time);
                } else if (ct == Types.TIMESTAMP) {
                    Timestamp time = r.getTimestamp(i + 1);
                    strValue = r.wasNull() ? null : getDatabase().getDateTimeFormatter().format(time);
                } else {
                    strValue = r.getString(i + 1);
                }
                row.setValue(i, strValue);
            }
            _rows.add(row);
            row.setStatus(Row.SYNC);
            if (rows++ > _database.getMaximumRowsToRetrieve()) {
                _incomplete = true;
                Log.message("Resultset row count exceeded maximum rows setting: %d", _database.getMaximumRowsToRetrieve());
                break;
            }
        }
    }


    public Table getTable() {
        return _table;
    }

    public String getTableName() {
        return null == _table ? _tableName : _table.getName();
    }

    public boolean isSyncable() {
        return _table != null;
    }

    public Database getDatabase() {
        return _database;
    }

    public void setDatabase(Database database) {
        _database = database;
    }

    public Object[] getParams() {
        return _params;
    }

    public void setParams(Object[] params) {
        _params = params;
    }

    public void newRow() {
        _rows.add(new Row(_columnNames.length));
    }

    public void synchronize() throws Exception {
        boolean autoCommit = _database.isAutoCommit();
        if (isSyncable()) {
            try {
                _database.setAutoCommit(false);
                // deletes go first
                for (Row row : _deletedRows) {
                    syncDelete(_table, row);
                }
                for (Row row : _rows) {
                    int status = row.getStatus();
                    if (status == Row.UPDATED) {
                        syncUpdate(_table, row);
                    } else if (status == Row.NEW) {
                        syncNew(_table, row);
                    }
                }

                _database.getCommander().commit();
                // update status
                _deletedRows.clear();
                for (Row row : _rows) {
                    row.setStatus(Row.SYNC);
                }
            } catch (Exception e) {
                if (e instanceof SQLException) {
                    _database.getCommander().rollback();
                }
                throw e;
            } finally {
                _database.setAutoCommit(autoCommit);
            }
        }
    }

    private void syncDelete(Table t, Row row) throws Exception {
        ArrayList<Object> params = new ArrayList<Object>();
        PrimaryKey[] pks = t.getPrimaryKeys();
        StringBuffer sb = new StringBuffer();
        sb.append("DELETE FROM ").append(t.getQuotedName()).append(" WHERE ");
        for (int i = 0; i < pks.length; i++) {
            PrimaryKey pk = pks[i];
            sb.append(pk.getColumnName()).append("=?");
            params.add(row.getValue(pk.getColumnName()));
            if (i + 1 < pks.length) {
                sb.append(" AND ");
            }
        }
        t.getDatabase().getCommander().execute(sb.toString(), params.toArray());
    }

    private void syncNew(Table t, Row row) throws Exception {
        StringBuffer sb = new StringBuffer();
        StringBuffer sb1 = new StringBuffer();
        ArrayList<Object> params = new ArrayList<Object>();
        sb.append("INSERT INTO ").append(t.getQuotedName()).append(" ( ");
        Column[] columns = t.getColumns();
        for (int i = 0; i < columns.length; i++) {
            Column column = columns[i];
            Object value = row.getValue(i);
            if (column.isPrimaryKey() && !Utils.hasValue(value)) {
                // skip possible autoincrement columns
                continue;
            }
            sb.append(quote(t.getDatabase(), column.getName()));
            sb1.append("? ");
            if (i + 1 < columns.length) {
                sb.append(", ");
                sb1.append(", ");
            }
            value = SQLCommander.sureType(value, column.getType(), getDatabase());
            params.add(value);
        }
        sb.append(" ) VALUES (").append(sb1).append(" )");
        t.getDatabase().getCommander().execute(sb.toString(), params.toArray());
    }


    private void syncUpdate(Table t, Row row) throws Exception {
        ArrayList<Object> params = new ArrayList<Object>();
        StringBuffer sb = new StringBuffer();
        Column[] columns = t.getColumns();
        sb.append("UPDATE ").append(t.getQuotedName()).append(" SET ");
        int cols = 0;
        for (int i = 0; i < columns.length; i++) {
            Column c = columns[i];
            if (!c.isPrimaryKey()) {
                Object ov = row.getOriginalValue(i);
                Object value = row.getValue(i);
                if ((null == ov && value != null) || (null != ov && !ov.equals(value))) {
                    if (cols > 0) {
                        sb.append(", ");
                    }
                    sb.append(quote(t.getDatabase(), c.getName())).append("=?");
                    params.add(SQLCommander.sureType(value, c.getType(), getDatabase()));
                    cols++;
                }
            }
        }
        PrimaryKey[] pks = t.getPrimaryKeys();
        sb.append(" WHERE ");
        for (int i = 0; i < pks.length; i++) {
            PrimaryKey pk = pks[i];
            sb.append(quote(t.getDatabase(), pk.getColumnName())).append("=?");
            Object value = row.getValue(pk.getColumnName());
            params.add(value);
            if (i + 1 < pks.length) {
                sb.append(" AND ");
            }
        }
        t.getDatabase().getCommander().execute(sb.toString(), params.toArray());
    }

    private String quote(Database database, String name) {
        return null == database ? name : database.quote(name);
    }

    public void deleteRow(int index) {
        Row r = _rows.get(index);
        _rows.remove(index);
        if (r.getStatus() != Row.NEW) {
            _deletedRows.add(r);
            r.setStatus(Row.DELETED);
        }
    }

    public Column[] getColumns() {
        return _columns;
    }

    public Column getColumn(int index) {
        return _columns[index];
    }

    public String getStatement() {
        return _statement;
    }

    public void setStatement(String statement) {
        _statement = statement;
    }

    public String getColumnName(int column) {
        return _columnNames[column];
    }

    public int getRowCount() {
        return _rows.size();
    }

    public int getColumnCount() {
        return _columnNames.length;
    }

    public boolean isCellEditable(int rowIndex, int columnIndex) {
        if (isEditable()) {
            Row row = getRow(rowIndex);
            return null != _table
                    && (row.getStatus() == Row.NEW || !_table.getColumns()[columnIndex].isPrimaryKey());
        } else {
            return false;
        }
    }

    public void setValueAt(Object aValue, int rowIndex, int columnIndex) {
        ResultModel.Row row = getRow(rowIndex);
        if (null != row) {
            row.setValue(columnIndex, aValue);
            fireTableCellUpdated(rowIndex, columnIndex);
        } else {
            Utils.debug("Invalid row index = " + rowIndex);
        }
    }

    public Object getValueAt(int rowIndex, int columnIndex) {
        Row r = getRow(rowIndex);
        return null == r ? null : r.getValue(columnIndex);
    }

    private Row getRow(int rowIndex) {
        return (rowIndex >= 0 && rowIndex < getRowCount() ? _rows.get(rowIndex) : null);
    }

    public ResultModel refresh() throws Exception {
        List<ISQLResult> rms = _database.getCommander().execute(getStatement(), getParams());
        if (rms.size() == 1 && rms.get(0) instanceof ResultModel) {
            ResultModel resultModel = (ResultModel) rms.get(0);
            resultModel.setTable(_table);
            return resultModel;
        }
        throw new IllegalStateException("Wrong resultset list size!");
    }

    public String toString() {
        return "ResultSet: " + getRowCount() + " row(s)";
    }

    public void setTable(Table table) {
        _table = table;
    }

    public Column getColumn(String name) {
        Integer colIndex = _nameMap.get(name);
        return null == colIndex ? null : _columns[colIndex];
    }

    private class Row {
        private Object[] _columns;
        private Object[] _orginals;
        private int _status = NEW;
        public static final int NEW = 0;
        public static final int UPDATED = 1;
        public static final int DELETED = 2;
        public static final int SYNC = 3;

        public Row(int n) {
            _columns = new Object[n];
        }

        public int getStatus() {
            return _status;
        }

        public void setStatus(int status) {
            _status = status;
            if (_status == SYNC) {
                backupRow();
            }
        }

        public Object getValue(String columnName) {
            Integer o = _nameMap.get(columnName);
            return getValue(o);
        }

        public Object getOriginalValue(int index) {
            return _orginals[index];
        }

        public void setValue(int index, Object v) {
            Object o = _columns[index];
            if (o == null || !o.equals(v)) {
                _columns[index] = v;
                if (getStatus() == SYNC) {
                    setStatus(UPDATED);
                }
            }
        }

        private void backupRow() {
            _orginals = new Object[_columns.length];
            System.arraycopy(_columns, 0, _orginals, 0, _columns.length);
        }

        public Object getValue(int columnIndex) {
            return _columns[columnIndex];
        }

        public Object[] getArray() {
            return _columns;
        }
    }
}
