package dbhelp.plugin.idea;

import bean.PropertyChangeMultiCaster;
import com.intellij.openapi.components.ProjectComponent;
import com.intellij.openapi.editor.Document;
import com.intellij.openapi.editor.EditorFactory;
import com.intellij.openapi.editor.LogicalPosition;
import com.intellij.openapi.editor.event.CaretEvent;
import com.intellij.openapi.editor.event.CaretListener;
import com.intellij.openapi.editor.event.DocumentListener;
import com.intellij.openapi.editor.event.EditorEventMulticaster;
import com.intellij.openapi.fileEditor.*;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.openapi.wm.ToolWindow;
import static com.intellij.openapi.wm.ToolWindowAnchor.BOTTOM;
import static com.intellij.openapi.wm.ToolWindowAnchor.LEFT;
import com.intellij.openapi.wm.ToolWindowManager;
import dbhelp.db.Procedure;
import dbhelp.db.Table;
import dbhelp.db.model.search.provider.TableAndProcedureSearchProvider;
import dbhelp.plugin.DBBrowserPane;
import dbhelp.plugin.OutputPane;
import dbhelp.plugin.action.portable.PopupMenuManager;
import dbhelp.plugin.complete.AbstractCompleteWindow;
import dbhelp.plugin.idea.action.FindTableInDiagram;
import dbhelp.plugin.idea.action.ViewProcedureSyntax;
import dbhelp.plugin.idea.action.ViewTablesDiagram;
import dbhelp.plugin.idea.complete.CompleteWindow;
import dbhelp.plugin.idea.complete.DocListener;
import dbhelp.plugin.idea.filetype.DiagramEditor;
import dbhelp.plugin.idea.filetype.DiagramFileType;
import dbhelp.plugin.idea.ui.IDEADiagram;
import dbhelp.plugin.idea.utils.IDEAUtils;
import org.jetbrains.annotations.NotNull;
import utils.DataCollector;
import utils.Icons;
import utils.Utils;

import javax.swing.*;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.util.HashMap;
import java.util.Map;

/**
 * User: huojun
 * Date: 12-Sep-2006
 * Time: 11:01:36 AM
 */
public class ProjectMain implements ProjectComponent {
    private static Map<Project, ProjectMain> _context = new HashMap<Project, ProjectMain>();
    private Project _project;
    private DBBrowserPane _browser;
    private OutputPane _outputPane;
    private PopupMenuManager _popupMenuManager = new PopupMenuManager();
    private AbstractCompleteWindow _completeWindow = new CompleteWindow();
    private DocumentListener _docListener = new DocListener();
    private DataCollector _dataCollector = new DataCollector();


    public ProjectMain(Project project) {
        _project = project;
    }

    public static ProjectMain getInstance() {
        return _context.get(IDEAUtils.getProject());
    }

    public PopupMenuManager getPopupMenuManager() {
        return _popupMenuManager;
    }

    public void projectOpened() {
        _context.put(_project, this);
        initToolWindow();
        FileDocumentManager fdm = FileDocumentManager.getInstance();
        fdm.addFileDocumentManagerListener(new MyFileDocumentManagerAdapter());
        javax.swing.FocusManager fm = FocusManager.getCurrentManager();
        fm.addPropertyChangeListener("permanentFocusOwner", new MyPropertyChangeListener());
        EditorEventMulticaster c = EditorFactory.getInstance().getEventMulticaster();
        c.addCaretListener(new MyCaretListener());
        FileEditorManager fem = FileEditorManager.getInstance(_project);
        fem.addFileEditorManagerListener(new MyFileEditorManagerAdapter());
        addIDEAActions(_popupMenuManager);
        _dataCollector.addListener(new TableAndProcedureSearchProvider());
        PropertyChangeMultiCaster.getInstance().addListener(OutputPane.class, new PropertyChangeListener() {
            public void propertyChange(PropertyChangeEvent evt) {
                if( _outputPane.equals(evt.getSource()) && "visible".equals(evt.getPropertyName())) {
                    IDEAUtils.showToolWindow(OutputPane.OUTPUT_WINDOW_ID);
                }
            }
        });
    }

    private void addIDEAActions(PopupMenuManager m) {
        m.addAction(new FindTableInDiagram(), Table.class);
        m.addAction(new ViewProcedureSyntax(), Procedure.class);
        m.addAction(new ViewTablesDiagram(), Table[].class);
    }

    public AbstractCompleteWindow getCompleteWindow() {
        return _completeWindow;
    }

    public void projectClosed() {
        unregisterToolWindow();
    }

    public void initComponent() {
    }

    public void disposeComponent() {
    }

    @NotNull
    public String getComponentName() {
        return "DBHelper";
    }

    public Project getProject() {
        return _project;
    }

    public OutputPane getOutputPane() {
        return _outputPane;
    }

    public DBBrowserPane getBrowser() {
        return _browser;
    }

    private void initToolWindow() {
        ToolWindowManager twm = ToolWindowManager.getInstance(_project);
        _browser = new DBBrowserPane();
        ToolWindow myToolWindow = twm.registerToolWindow(DBBrowserPane.BROWSER_WINDOW_ID, _browser, LEFT);
        myToolWindow.setTitle("DB Browser");
        Icons icons = Icons.getInstance();
        myToolWindow.setIcon(icons.get(0));
        _outputPane = new OutputPane();

        myToolWindow = twm.registerToolWindow(OutputPane.OUTPUT_WINDOW_ID, _outputPane, BOTTOM);
        myToolWindow.setIcon(icons.get(29));
        myToolWindow.setTitle("Output");


    }

    private void unregisterToolWindow() {
        ToolWindowManager twm = ToolWindowManager.getInstance(_project);
        twm.unregisterToolWindow(DBBrowserPane.BROWSER_WINDOW_ID);
        twm.unregisterToolWindow(OutputPane.OUTPUT_WINDOW_ID);
    }

    public DataCollector getDataCollector() {

        return _dataCollector;
    }


    private class MyFileDocumentManagerAdapter extends FileDocumentManagerAdapter {
        public void fileContentLoaded(VirtualFile virtualFile, Document document) {
            String name = virtualFile.getName();
            if (name.endsWith(".sql")) {
                document.addDocumentListener(_docListener);
            }
        }
    }

    private static class MyPropertyChangeListener implements PropertyChangeListener {
        public void propertyChange(PropertyChangeEvent evt) {
            Object newValue = evt.getNewValue();
            if (newValue instanceof JComponent) {
                JComponent c = (JComponent) newValue;
                if (null != IDEAUtils.getProject()) {
                    AbstractCompleteWindow cw = CompleteWindow.getInstance();
                    if (!cw.isAncestorOf(c)) {
                        cw.closeCompleter();
                    }
                }
            }
        }
    }

    private static class MyCaretListener implements CaretListener {
        public void caretPositionChanged(CaretEvent e) {
            FileDocumentManager fdm = FileDocumentManager.getInstance();
            VirtualFile file = fdm.getFile(e.getEditor().getDocument());
            if (null != file) {
                String name = file.getName();
                if (name.endsWith(".sql")) {
                    LogicalPosition np = e.getNewPosition();
                    LogicalPosition op = e.getOldPosition();
                    if (np.column - op.column > 1 || np.line != op.line) {
                        CompleteWindow.getInstance().closeCompleter();
                    }
                }
            }
        }
    }

    private static class MyFileEditorManagerAdapter extends FileEditorManagerAdapter {
        public void fileOpened(final FileEditorManager source, final VirtualFile file) {
            if (file.getFileType() instanceof DiagramFileType) {
                SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        loadDiagram(source, file);
                    }
                });
            }
        }

        private void loadDiagram(FileEditorManager source, VirtualFile file) {
            FileEditor[] editors = source.getEditors(file);
            for (FileEditor editor : editors) {
                if (editor instanceof DiagramEditor) {
                    DiagramEditor de = (DiagramEditor) editor;
                    File f = new File(file.getPath());
                    if (f.exists() && f.length() > 0) {
                        IDEADiagram diagram = de.getDiagram();
                        diagram.setFile(f);
                        try {
                            diagram.load();
                        } catch (Exception e) {
                            Utils.error(e);
                        }
                    }
                    return;
                }
            }
        }
    }

}
