package dbhelp.plugin.idea;

import bean.PropertyChangeMultiCaster;
import bean.editor.FileHolder;
import com.intellij.ide.highlighter.custom.CustomFileHighlighter;
import com.intellij.ide.highlighter.custom.CustomFileTypeLexer;
import com.intellij.ide.highlighter.custom.SyntaxTable;
import com.intellij.ide.highlighter.custom.impl.CustomFileType;
import com.intellij.ide.structureView.StructureViewBuilder;
import com.intellij.lexer.Lexer;
import com.intellij.lexer.LexerPosition;
import com.intellij.openapi.components.ApplicationComponent;
import com.intellij.openapi.editor.colors.TextAttributesKey;
import com.intellij.openapi.fileTypes.FileType;
import com.intellij.openapi.fileTypes.FileTypeManager;
import com.intellij.openapi.fileTypes.SyntaxHighlighter;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.openapi.options.Configurable;
import com.intellij.openapi.options.ConfigurationException;
import com.intellij.openapi.actionSystem.AnAction;
import com.intellij.openapi.actionSystem.ActionManager;
import static com.intellij.psi.CustomHighlighterTokenType.*;
import com.intellij.psi.tree.IElementType;
import dbhelp.db.Database;
import dbhelp.plugin.idea.highlight.SQLSyntaxTable;
import dbhelp.plugin.idea.action.CompleteAction;
import dbhelp.plugin.idea.action.SmartCompleteAction;
import dbhelp.plugin.idea.action.EASelectAll;
import dbhelp.plugin.ConfigPane;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nls;
import org.jetbrains.annotations.Nullable;
import org.jetbrains.annotations.NonNls;
import template.TemplateManager;
import utils.Config;
import utils.Icons;
import utils.SingletonProvider;
import utils.Utils;

import javax.swing.*;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.InputStream;
import java.lang.reflect.Method;
import java.net.URL;
import java.net.URLDecoder;
import java.util.*;
import java.util.zip.ZipEntry;
import java.util.jar.JarFile;
import java.util.jar.JarEntry;

/**
 * User: huojun
 * Date: 21-Sep-2006
 * Time: 10:31:14 AM
 */
public class SQLFileTypeSupport implements ApplicationComponent, Configurable {
    public static void main(String[] args) throws Exception {
        SQLFileTypeSupport s = new SQLFileTypeSupport();
        s.unzipTemplates();
    }

    public SQLFileTypeSupport() {
        try {
            SingletonProvider.createProvider("dbhelp.plugin.idea.utils.IDEASingletonProvider");
        } catch (Throwable throwable) {
            throw new RuntimeException(throwable);
        }
    }

    public void initComponent() {
        restoreConfigBean();
        SQLFileType type = new SQLFileType(new SQLSyntaxTable());
        type.setDescription("SQL Script File");
        type.setName("SQL");
        type.setIcon(Icons.getInstance().get(0));
        FileTypeManager.getInstance().registerFileType(type, new String[]{"sql"});
        TemplateManager.initialize();
        // Intercept CodeCompletion, because by default it is disabled
        interceptAction("CodeCompletion", new CompleteAction());
        interceptAction("SmartTypeCompletion", new SmartCompleteAction());
        interceptAction("$SelectAll", new EASelectAll());
    }

    private void interceptAction(String actionId, AnAction newAction) {
        ActionManager am = ActionManager.getInstance();
        AnAction action = am.getAction(actionId);
        if (null != action) {
            newAction.copyFrom(action);
        }
        am.unregisterAction(actionId);
        am.registerAction(actionId, newAction);
    }

    private void restoreConfigBean() {
        File configFile = Utils.getConfigFile();
        if (configFile.exists()) {
            try {
                Config.restore(configFile);
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
        try {
            installTemplates();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    private void installTemplates() throws Exception {
        Config config = Config.getConfig();
        if (config.isUseInternalTemplates()) {
            System.out.println("Use internal templates...");
            File file = unzipTemplates();
            System.out.println("Templates are located at: " + file);
            if (null != file) {
                config.setTemplateFileFolder(new FileHolder(file));
            }
        }
    }

    private File unzipTemplates() throws Exception {
        String jarFile = getJarFileName("resources/templates/templates.xml");
        if (null != jarFile) {
            File templateFolder = Utils.createDBHelperFolder("templates");
            if (templateFolder.exists()) {
                JarFile jf = new JarFile(jarFile);
                Enumeration<JarEntry> entries = jf.entries();
                while (entries.hasMoreElements()) {
                    JarEntry entry = entries.nextElement();
                    String name = entry.getName();
                    String prefix = "resources/templates/";
                    if (name.startsWith(prefix) && !entry.isDirectory()) {
                        ZipEntry zipEntry = jf.getEntry(name);
                        System.out.println("Unzip: " + name);
                        int nameIndex = name.lastIndexOf('/');
                        String path = nameIndex > prefix.length() ? name.substring(prefix.length(), nameIndex) : "";
                        File fileFolder = Utils.mkdirs(templateFolder, path);
                        if (fileFolder.exists()) {
                            File templateFile = new File(fileFolder, name.substring(nameIndex + 1));
                            System.out.println("To: " + templateFile);
                            InputStream inputStream = jf.getInputStream(zipEntry);
                            Utils.createFile(templateFile, new InputStreamReader(inputStream));
                            inputStream.close();
                        } else {
                            System.out.println(fileFolder + " is bad name!");
                        }
                    }
                }
            } else {
                throw new RuntimeException("Failed to create folder: " + templateFolder);
            }
            return templateFolder;
        }
        return null;
    }

    private String getJarFileName(String resourceURL) {
        try {
            ClassLoader cl = getClass().getClassLoader();
            URL url = cl.getResource(resourceURL);
            String str = url.toString();
            String prefix = "file:/";
            String path = str.substring(str.indexOf(prefix) + prefix.length(), str.indexOf('!'));
            String jarFile = URLDecoder.decode(path, "UTF-8");
            Utils.debug("f = " + jarFile);
            return jarFile;
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }
    }

    public void disposeComponent() {
        Database[] databases = Config.getConfig().getDatabases();
        for (int i = 0; null != databases && i < databases.length; i++) {
            Database d = databases[i];
            d.disconnect();
        }
        try {
            Config.getConfig().store(Utils.getConfigFile());
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    @NotNull
    public String getComponentName() {
        return "SQLFileTypeSupport";
    }

    @Nls
    public String getDisplayName() {
        return "DBHelper";
    }

    public Icon getIcon() {
        return Icons.getInstance().get(25);
    }

    @Nullable
    @NonNls
    public String getHelpTopic() {
        return "";
    }

    public JComponent createComponent() {
        return new ConfigPane();
    }

    public boolean isModified() {
        return false;
    }

    public void apply() throws ConfigurationException {

    }

    public void reset() {

    }

    public void disposeUIResources() {

    }

    public static class SQLFileType implements FileType {
        private CustomFileType _delegate;
        private SQLSyntaxHighlighter _highlighter;

        public SQLFileType(SyntaxTable syntaxTable) {
            _delegate = new CustomFileType(syntaxTable);
            _delegate.initSupport();
        }


        @NotNull
        public String getName() {
            return _delegate.getName();
        }

        @NotNull
        public String getDescription() {
            return _delegate.getDescription();
        }

        public void setName(String s) {
            _delegate.setName(s);
        }

        public void setDescription(String s) {
            _delegate.setDescription(s);
        }

        @NotNull
        public String getDefaultExtension() {
            return "sql";
        }

        public Icon getIcon() {
            return _delegate.getIcon();
        }

        public boolean isBinary() {
            return _delegate.isBinary();
        }

        public boolean isReadOnly() {
            return _delegate.isReadOnly();
        }

        public String getCharset(@NotNull VirtualFile virtualFile) {
            return _delegate.getCharset(virtualFile);
        }

        public void setIcon(Icon icon) {
            _delegate.setIcon(icon);
        }

        public StructureViewBuilder getStructureViewBuilder(@NotNull VirtualFile virtualFile, @NotNull Project project) {
            return _delegate.getStructureViewBuilder(virtualFile, project);
        }

        @NotNull
        public SyntaxHighlighter getHighlighter(Project project) {
            if (null == _highlighter) {
                try {
                    Class<? extends CustomFileType> aClass = _delegate.getClass();
                    Method getHighlighterMethod = aClass.getMethod("getHighlighter", Project.class);
                    CustomFileHighlighter syntaxHighlighter = (CustomFileHighlighter) getHighlighterMethod.invoke(_delegate, project);
                    _highlighter = new SQLSyntaxHighlighter(syntaxHighlighter);
                } catch (Exception e) {
                    Utils.debug("SQLFileTypeSupport$SQLFileType.getHighlighter: " + e);
                }
            }
            return _highlighter;
        }

        @NotNull
        public SyntaxHighlighter getHighlighter(Project project, VirtualFile vf) {
            if (null == _highlighter) {
                _highlighter = new SQLSyntaxHighlighter((CustomFileHighlighter) _delegate.getHighlighter(project, vf));
            }
            return _highlighter;
        }
    }

    private static class SQLSyntaxHighlighter implements SyntaxHighlighter {
        private CustomFileHighlighter _highlighter;
        private SQLLexer _lexer;

        public SQLSyntaxHighlighter(CustomFileHighlighter highlighter) {
            _highlighter = highlighter;
        }

        @NotNull
        public Lexer getHighlightingLexer() {
            if (null == _lexer) {
                _lexer = new SQLLexer((CustomFileTypeLexer) _highlighter.getHighlightingLexer());
            }
            return _lexer;
        }

        @NotNull
        public TextAttributesKey[] getTokenHighlights(IElementType iElementType) {
            return _highlighter.getTokenHighlights(iElementType);
        }

        public static Map getKeys() {
            return CustomFileHighlighter.getKeys();
        }
    }

    public static class SQLLexer implements Lexer {
        private CustomFileTypeLexer _lexer;
        private Set<String> _keywords = new HashSet<String>();

        public void setDatabase(Database database) {
            Utils.debug("SQLLexer db chagne: " + database);
            String[] keywords = database.getKeywords();
            _keywords.clear();
            if (null != keywords) {
                _keywords.addAll(Arrays.asList(keywords));
            } else {
                _keywords.addAll(Database.SQL92_KEYWORD_LIST);
            }
            PropertyChangeMultiCaster.getInstance().fireChangeEvent(this, "database");
        }

        public Set<String> getKeywords() {
            return _keywords;
        }

        public SQLLexer(CustomFileTypeLexer lexer) {
            _lexer = lexer;
        }

        public void start(char[] chars) {
            _lexer.start(chars);
        }

        public void start(char[] chars, int i, int i1) {
            _lexer.start(chars, i, i1);
        }

        public void start(char[] chars, int i, int i1, int i2) {
            _lexer.start(chars, i, i1, i2);
        }

        public int getState() {
            return _lexer.getState();
        }

        public IElementType getTokenType() {
            IElementType type = _lexer.getTokenType();
            char[] chars = _lexer.getBuffer();
            int start = _lexer.getTokenStart();
            int count = _lexer.getTokenEnd() - _lexer.getTokenStart();
            if (chars != null && start >= 0 && start + count <= chars.length) {
                try {
                    String s = new String(chars, start, count);
                    if (type == IDENTIFIER) {
                        //                Utils.debug("s = " + s);
                        if (_keywords.contains(s.toUpperCase())) {
                            //                    Utils.debug("Corrected keyword = " + s);
                            return KEYWORD_1;
                        }
                    } else if (type == STRING) {
                        if (s.length() > 0 && s.charAt(0) == '"') {
                            return KEYWORD_4;
                        }
                    }
                } catch (Exception e) {                    
                    Utils.debug(e);
                }
            }
            return type;
        }

        public int getTokenStart() {
            return _lexer.getTokenStart();
        }

        public int getTokenEnd() {
            return _lexer.getTokenEnd();
        }

        public void advance() {
            _lexer.advance();
        }

        public char[] getBuffer() {
            return _lexer.getBuffer();
        }

        public int getBufferEnd() {
            return _lexer.getBufferEnd();
        }

        public LexerPosition getCurrentPosition() {
            return _lexer.getCurrentPosition();
        }

        public void restore(LexerPosition position) {
            _lexer.restore(position);
        }
    }

}
