package utils.text;

/**
 * User: David Huo
 * Date: Sep 19, 2006
 * Time: 3:10:30 PM
 */
public class ForwardsEnumerator implements IWordEnumeration {
    private int _index;
    private int _offset;
    private CharSequence _text;
    private TYPE _type = TYPE.NULL;
    private String _operators = WordScanner.DEFAULT_OPERATORS;


    public ForwardsEnumerator() {
    }

    public ForwardsEnumerator(CharSequence text, int index) {
        _index = index;
        _text = text;
        _offset = _index;
    }


    public CharSequence getText() {
        return _text;
    }

    public void setText(CharSequence text) {
        _index = _offset = 0;
        _text = text;
    }

    public String getOperators() {
        return _operators;
    }

    public int getLine() {
        int line = 0;
        String text = _text == null ? "" : _text.toString();
        int index = text.indexOf('\n');
        System.out.println("text = " + text);
        while( index >= 0 ) {
            if( index >= _index ) {
                return line;
            }
            line++;
            index = text.indexOf('\n', index+1);
        }
        return line;
    }

    public void setOperators(String operators) {
        _operators = operators;
    }

    public boolean hasMoreElements() {
        return null != _text && _index < _text.length();
    }

    private char peek(int offset) {
        return offset < _text.length() ? _text.charAt(offset) : 0;
    }

    public CharSequence nextElement() {
        CharSequence token = nextToken();
        _offset = _index - token.length();
        return token;
    }

    private CharSequence nextToken() {
        char c = peek(_index);
        if( c == '\r' && peek(_index+1) == '\n') {
            return consumeDOSNewLine(_index);
        } else if (c == '\n') {
            return consumeNewLine(_index);
        } else if ((Character.isDigit(c) || c == '.')
                && Character.isDigit(peek(_index + 1))) {
            return consumeNumber(_index);
        } else if (c == '-' && peek(_index + 1) == '-') {
            return consumeLineComment(_index);
        } else if (c == '{') {
            return consumeComment1(_index);
        } else if (c == '\'') {
            return consumeString(_index);
        } else if (Character.isWhitespace(c)) {
            return consumeWS(_index);
        } else if (isCommentBegin(_index)) {
            return consumeComment(_index);
        } else if (isOperator(_index)) {
            return consumeOperator(_index);
        } else if ( c == '"' || c == '`') {
            return consumeObjectName(_index, c);
        } else {
            return consumeChars(_index);
        }
    }

    private CharSequence consumeComment1(int index) {
        int p = index + 1;
        int len = _text.length();
        while (p < len) {
            char c = peek(p);
            if (c == '}') {
                break;
            }
            p++;
        }
        _index = p < len ? p + 1 : p;
        _type = TYPE.MULTILINE_COMMENT;
        return _text.subSequence(index, _index);
    }

    private CharSequence consumeNumber(int index) {
        char first = peek(index);
        boolean hasDot = first == '.';
        int p = index + 1;
        int len = _text.length();
        while (p < len) {
            char c = peek(p);
            if (!Character.isDigit(c)) {
                if( hasDot ) {
                    break;
                } else if( c == '.' ) {
                    hasDot = true;
                } else {
                    break;
                }
            }
            p++;
        }
        _index = p;
        _type = TYPE.NUMBER;
        return _text.subSequence(index, _index);
    }

    private CharSequence consumeObjectName(int index, char quoteChar) {
        int p = index + 1;
        int len = _text.length();
        while (p < len) {
            char c = peek(p);
            if (c == quoteChar) {
                break;
            }
            p++;
        }
        _index = p < len ? p + 1 : p;
        _type = TYPE.OBJECT;
        return _text.subSequence(index, _index);
    }

    private CharSequence consumeLineComment(int index) {
        int p = index + 2;
        int len = _text.length();
        while( p<len ) {
            char c = peek(p);
            if( c == '\n' ) {
                break;
            }
            p++;
        }
        _index = p;
        _type = TYPE.SINGLELINE_COMMENT;
        return _text.subSequence(index, p);
    }


    private CharSequence consumeString(int index) {
        int p = index + 1;
        int len = _text.length();
        while (p < len) {
            char c = peek(p);
            if (c == '\'') {
                if (peek(p + 1) == '\'') {
                    p += 1;
                } else {
                    break;
                }
            }
            p++;
        }
        _index = p < len ? p + 1 : p;
        _type = TYPE.STRING;
        return _text.subSequence(index, _index);
    }

    public int getIndex() {
        return _index;
    }

    public void setIndex(int index) {
        _index = index;
    }

    public TYPE getType() {
        return _type;
    }

    public int getOffset() {
        return _offset;
    }

    private CharSequence consumeDOSNewLine(int index) {
        _index+=2;
        _type = TYPE.NEWLINE;
        return _text.subSequence(index, index + 2);
    }

    private CharSequence consumeNewLine(int index) {
        _index++;
        _type = TYPE.NEWLINE;
        return _text.subSequence(index, index + 1);
    }

    private CharSequence consumeChars(int index) {
        int p = index;
        int len = _text.length();
        while (p < len) {
            char c = peek(p);
            if (Character.isWhitespace(c)
                    || isOperator(p)
                    || isCommentBegin(p)
                    ) {
                break;
            } else {
                p++;
            }
        }
        _index = p;
        CharSequence cs = _text.subSequence(index, p);
        _type = WordScanner.isNumber(cs) ? TYPE.NUMBER : TYPE.WORD;
        return cs;
    }

    private CharSequence consumeComment(int index) {
        int p = index;
        int len = _text.length();
        while (p < len) {
            char c = peek(p);
            char c1 = peek(p + 1);
            if (c == '*' && c1 == '/') {
                p = p + 2;
                break;
            }
            p++;
        }
        _index = p;
        _type = TYPE.MULTILINE_COMMENT;
        return _text.subSequence(index, p);
    }

    private boolean isCommentBegin(int index) {
        char c = peek(index);
        char c1 = peek(index + 1);
        return c == '/' && c1 == '*' || c == '{';
    }

    private CharSequence consumeOperator(int index) {
        char c = peek(index);
        char c1 = peek(index + 1);
        int n = ((c == '|' && c1 == '|')
                || (c == '&' && c1 == '&')
                || (c == '+' && c1 == '+')
                || (c == '=' && c1 == '+')
                || (c == '=' && c1 == '-')
                || (c == '=' && c1 == '%')
                || (c == '=' && c1 == '*')
                || (c == '=' && c1 == '/')
                || (c == '*' && c1 == '/')
                || (c == '/' && c1 == '*')
                || (c == '-' && c1 == '-')
        ) ? 1 : 0;
        _index = _index + (n + 1);
        _type = TYPE.OPERATOR;
        return _text.subSequence(index, _index);
    }


    private boolean isOperator(int index) {
        char c = peek(index);
        return _operators.indexOf(c) >= 0;
    }

    private CharSequence consumeWS(int index) {
        int p = index;
        int len = _text.length();
        while (p < len) {
            char c = peek(p);
            if (c == '\n' || !Character.isWhitespace(c)) {
                break;
            } else {
                p++;
            }
        }
        _index = p;
        _type = TYPE.WHITESPACES;
        return _text.subSequence(index, p);
    }
}
