package utils.text.editor;

import dbhelp.db.Database;
import utils.text.ForwardsEnumerator;
import utils.text.IWordEnumeration;
import static utils.text.IWordEnumeration.TYPE.*;

import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.event.UndoableEditEvent;
import javax.swing.event.UndoableEditListener;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;
import javax.swing.text.JTextComponent;
import javax.swing.text.PlainDocument;
import javax.swing.undo.UndoManager;
import java.awt.*;
import java.util.*;
import java.util.List;

/**
 * User: huojun
 * Date: 18-Oct-2006
 * Time: 11:37:59 AM
 */
public class TokenDocument extends PlainDocument {
    protected List<TokenLine> _lines = new ArrayList<TokenLine>();
    protected Set<String> _keywords = new HashSet<String>();
    protected Map<IWordEnumeration.TYPE, RenderAttribute>
            _attributes = new HashMap<IWordEnumeration.TYPE, RenderAttribute>();
    protected ForwardsEnumerator _scanner = new ForwardsEnumerator();
    private MyDocumentListener _docListener = new MyDocumentListener();
    protected boolean _parserReady;
    protected UndoManager _undoMgr = new UndoManager();
    protected Set<Integer> _markedLines = new HashSet<Integer>();
    private Color _markedLineColor = Color.red;
    private JTextComponent _editor;


    public TokenDocument() {
        super();
        setup();
    }

    private void setup() {
        addDocumentListener(_docListener);
        Font f = new Font(null, Font.PLAIN, 12);
        addRenderAttribute(new RenderAttribute(KEYWORD, f, new Color(0, 0, 159)));
        addRenderAttribute(new RenderAttribute(STRING, f, new Color(92, 151, 89)));
        addRenderAttribute(new RenderAttribute(MULTILINE_COMMENT, f, Color.gray));
        addRenderAttribute(new RenderAttribute(SINGLELINE_COMMENT, f, Color.gray));
        addRenderAttribute(new RenderAttribute(OBJECT, f, new Color(128, 128, 64)));
        addRenderAttribute(new RenderAttribute(NUMBER, f, new Color(128, 64, 0)));
        addKeywords(Database.SQL92_KEYWORD_LIST);
        addUndoableEditListener(new UndoableEditListener() {
            public void undoableEditHappened(UndoableEditEvent e) {
                _undoMgr.addEdit(e.getEdit());
            }
        });
    }

    public TokenDocument(Content c) {
        super(c);
        setup();
    }


    public boolean isParserReady() {
        return _parserReady;
    }

    public void addRenderAttribute(RenderAttribute a) {
        _attributes.put(a.getType(), a);
    }


    public void addKeywords(String[] keywords) {
        _keywords.addAll(Arrays.asList(keywords));
    }

    private void parse(String text) {
        _parserReady = false;
        _lines.clear();
        _scanner.setText(text);
        if (text.length() > 0) {
            List<Token> line = new ArrayList<Token>();
            while (_scanner.hasMoreElements()) {
                CharSequence tokenText = _scanner.nextElement();
                IWordEnumeration.TYPE type = _scanner.getType();
                if (type == NEWLINE) {
                    addParsedLine(line);
                } else if (type == MULTILINE_COMMENT || type == STRING) {
                    String literal = tokenText.toString();
                    String[] strings = literal.split("\\n");
                    if (strings.length > 0) {
                        line.add(new Token(strings[0], type));
                        if (strings.length > 1) {
                            addParsedLine(line);
                            for (int i = 1; i < strings.length; i++) {
                                line.add(new Token(strings[i], type, true));
                                if (i + 1 < strings.length) {
                                    addParsedLine(line);
                                }
                            }
                        }
                    }
                } else {
                    Token t = new Token(tokenText, type);
                    line.add(t);
                }
            }
            if (line.size() > 0) {
                addParsedLine(line);
            }
        }
        _parserReady = true;
    }

    private void addParsedLine(List<Token> line) {
        TokenLine tl = new TokenLine(line.toArray(new Token[line.size()]));
        _lines.add(tl);
        line.clear();
    }

    public TokenLine getTokenLine(int line) {
        return line >= 0 && line < _lines.size() ? _lines.get(line) : null;
    }

    public boolean isKeyword(String s) {
        return _keywords.contains(s);
    }

    public RenderAttribute getRenderAttribute(IWordEnumeration.TYPE type) {
        return _attributes.get(type);
    }

    private void updateTokens(DocumentEvent e) {
        Document doc = e.getDocument();
//        Element root = getDefaultRootElement();
//        DocumentEvent.ElementChange ec = e.getChange(root);
//        if( ec != null ) {
//            System.out.println("line = " + ec.getIndex());
//        }
        try {
            parse(doc.getText(0, doc.getLength()));
        } catch (BadLocationException e1) {
            e1.printStackTrace();
        }
    }

    public void undo() {
        if( _undoMgr.canUndo() ) {
            _undoMgr.undo();
        }
    }

    public void markLine(int line) {
        _markedLines.add(line);

    }

    public boolean isMarkedLine(int line) {
        return _markedLines.contains(line);
    }

    public Color getMarkedLineColor() {
        return _markedLineColor;
    }

    public void clearMarkedLines() {
        _markedLines.clear();
    }

    public void setEditor(JTextComponent editor) {
        _editor = editor;
    }

    public JTextComponent getEditor() {
        return _editor;
    }

    public void addKeywords(Set<String> keywords) {
        _keywords.addAll(keywords);
    }

    private class MyDocumentListener implements DocumentListener {
        public void insertUpdate(DocumentEvent e) {
            updateTokens(e);
        }

        public void removeUpdate(DocumentEvent e) {
            updateTokens(e);
        }

        public void changedUpdate(DocumentEvent e) {
            updateTokens(e);
        }
    }
}
