package utils.text.editor;

import dbhelp.db.Database;
import dbhelp.plugin.complete.ISQLResolver;
import dbhelp.plugin.complete.utils.CompleterItem;
import utils.Utils;
import utils.text.WordScanner;
import utils.text.editor.complete.TokenCompleter;

import javax.swing.*;
import static javax.swing.KeyStroke.getKeyStroke;
import javax.swing.event.CaretEvent;
import javax.swing.event.CaretListener;
import javax.swing.event.DocumentEvent;
import static javax.swing.event.DocumentEvent.EventType;
import javax.swing.event.DocumentListener;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;
import javax.swing.text.PlainDocument;
import java.awt.*;
import static java.awt.Event.CTRL_MASK;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import static java.awt.event.KeyEvent.VK_SPACE;
import static java.awt.event.KeyEvent.VK_Z;
import java.util.Collection;
import java.util.Set;

/**
 * User: huojun
 * Date: 18-Oct-2006
 * Time: 11:57:59 PM
 */
public class TokenEditor extends JEditorPane {
    protected TokenCompleter _completer = new TokenCompleter();
    protected Database _database;
    protected static final String OLD_ACTION = "oldAction";
    protected MyDocumentListener _docListener = new MyDocumentListener();

    public static void main(String[] args) {
        JFrame f = new JFrame();
        Container c = f.getContentPane();
        c.setLayout(new BorderLayout());
        TokenEditor view = new TokenEditor();
        view.markLine(5);
        view.markLine(8);
        c.add(new JScrollPane(view), BorderLayout.CENTER);
        f.setSize(640, 480);
        f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        f.setVisible(true);
    }

    public TokenEditor() {
        setEditorKit(new TokenEditorKit("sql/text"));
        setFont(new Font(null, Font.BOLD, 12));
        TokenDocument doc = getTokenDocument();
        doc.setEditor(this);
        setTabSize(4);
        _completer.setEditor(this);

        addHotKey("undo", getKeyStroke(VK_Z, Event.CTRL_MASK), new ActionUndo());
        addHotKey("complete", getKeyStroke(VK_SPACE, CTRL_MASK), new ActionComplete());
        addHotKey("escape", getKeyStroke(KeyEvent.VK_ESCAPE, 0), new ActionEscape());

        overrideAction(getKeyStroke(KeyEvent.VK_UP, 0), new ActionUp());
        overrideAction(getKeyStroke(KeyEvent.VK_DOWN, 0), new ActionDown());
        overrideAction(getKeyStroke(KeyEvent.VK_ENTER, 0), new ActionEnter());
        overrideAction(getKeyStroke(KeyEvent.VK_LEFT, 0), new leftRightAction());
        overrideAction(getKeyStroke(KeyEvent.VK_RIGHT, 0), new leftRightAction());
        overrideAction(getKeyStroke(KeyEvent.VK_TAB, 0), new ActionTab());
        overrideAction(getKeyStroke(KeyEvent.VK_BACK_SPACE, 0), new ActionBackspace());
        overrideAction(getKeyStroke(KeyEvent.VK_A, CTRL_MASK), new ActionSelectAll());
        overrideAction(getKeyStroke(KeyEvent.VK_PAGE_UP, 0), new ActionPageUp());
        overrideAction(getKeyStroke(KeyEvent.VK_PAGE_DOWN, 0), new ActionPageDown());

        addCaretListener(new MyCaretListener());
        getDocument().addDocumentListener(_docListener);
    }

    public int getTabSize() {
        Object o = getDocument().getProperty(PlainDocument.tabSizeAttribute);
        return o instanceof Integer ? (Integer)o : 8;
    }
    public void setTabSize(int value) {
        getDocument().putProperty(PlainDocument.tabSizeAttribute, value);
    }

    public void addKeywords(String[] keywords) {
        getTokenDocument().addKeywords(keywords);
    }

    public void undo() {
        TokenDocument doc = getTokenDocument();
        doc.undo();
    }

    public boolean overrideAction(KeyStroke key, AbstractAction a) {
        Object actionName = getInputMap().get(key);
        if (actionName != null) {
//            Utils.debug("TokenEditor.overrideAction: " + actionName);
            ActionMap am = getActionMap();
            Action action = am.get(actionName);
            if (null != action) {
                a.putValue(OLD_ACTION, action);
                am.put(actionName, a);
                return true;
            }
        }
        return false;
    }

    public TokenCompleter getCompleter() {
        return _completer;
    }

    public Database getDatabase() {
        return _database;
    }

    public void setDatabase(Database database) {
        _database = database;
    }

    public void addHotKey(String id, KeyStroke key, Action a) {
        getInputMap().put(key, id);
        getActionMap().put(id, a);
    }

    private TokenDocument getTokenDocument() {
        return (TokenDocument) getDocument();
    }

    public void clearMarkedLines() {
        getTokenDocument().clearMarkedLines();
    }

    /**
     * The user needs to call repaint to show the marked lines
     *
     * @param line the line number to be marked
     */
    public void markLine(int line) {
        getTokenDocument().markLine(line);
    }

    public void addKeywords(Set<String> keywords) {
        getTokenDocument().addKeywords(keywords);
    }


    private class ActionComplete extends AbstractAction {
        public void actionPerformed(ActionEvent e) {
            try {
                Database database = getDatabase();
                if ( null != database ) {
                    if (!database.isConnected()) {
                        database.connect();
                    }
                    Document doc = getDocument();
                    CharSequence text = doc.getText(0, doc.getLength());
                    ISQLResolver resolver = database.getResolver();
                    int offset = getCaretPosition() - 1;
                    Collection<CompleterItem> completerItems = resolver.resolve(database, text, offset);
                    _completer.setText(database.toString());
                    _completer.setItems(completerItems);
                    CharSequence filter = WordScanner.getCompletingWord(text, offset);
                    _completer.setFilter(filter);
                    _completer.showCompleter();
                } else {
                    Utils.debug("Database is null");
                }
            } catch (IllegalStateException e1) {
                // ignore the wrong completion place error
            } catch (Exception e1) {
                Utils.error(e1);
            }
        }
    }

    private static abstract class OverrideAction extends AbstractAction {
        protected void defaultAction(ActionEvent e) {
            Action action = (Action) getValue(OLD_ACTION);
            if (null != action) {
                action.actionPerformed(e);
            }
        }

        protected TokenEditor getEditor(ActionEvent e) {
            return (TokenEditor) e.getSource();
        }


        public void actionPerformed(ActionEvent e) {
            if (getEditor(e).getCompleter().isVisible()) {
                overrideEvent(e);
            } else {
                defaultAction(e);
            }
        }

        protected abstract void overrideEvent(ActionEvent e);
    }

    private static class ActionUp extends OverrideAction {

        protected void overrideEvent(ActionEvent e) {
            getEditor(e).getCompleter().up();
        }
    }

    private static class ActionDown extends OverrideAction {

        protected void overrideEvent(ActionEvent e) {
            getEditor(e).getCompleter().down();
        }
    }

    private static class ActionEnter extends OverrideAction {

        protected void overrideEvent(ActionEvent e) {
            TokenCompleter c = getEditor(e).getCompleter();
            c.completeWord();
            c.closeCompleter();
        }
    }

    private class ActionBackspace extends OverrideAction {

        protected void overrideEvent(ActionEvent e) {
            if (_completer.isVisible()) {
                int pos = getEditor(e).getCaretPosition();
                if (pos > 0) {
                    try {
                        // try to test if dot will be deleted
                        String s = getDocument().getText(pos - 1, 1);
                        _docListener.setDeletingDot(".".equals(s));
                        if ("\n".equals(s) || " ".equals(s)) {
                            _completer.closeCompleter();
                        }
                    } catch (BadLocationException e1) {
                        Utils.debug("TokenEditor$ActionBackspace.overrideEvent: " + e1);
                    }
                }
            }
            defaultAction(e);
        }
    }

    private static class ActionSelectAll extends OverrideAction {

        protected void overrideEvent(ActionEvent e) {
            getEditor(e).getCompleter().selectAll();
        }
    }

    private static class ActionEscape extends OverrideAction {

        protected void overrideEvent(ActionEvent e) {
            getEditor(e).getCompleter().closeCompleter();
        }
    }

    private static class leftRightAction extends OverrideAction {

        protected void overrideEvent(ActionEvent e) {
            getEditor(e).getCompleter().closeCompleter();
            defaultAction(e);
        }
    }

    private static class ActionPageUp extends OverrideAction {

        protected void overrideEvent(ActionEvent e) {
            getEditor(e).getCompleter().pageUp();
        }
    }

    private static class ActionPageDown extends OverrideAction {

        protected void overrideEvent(ActionEvent e) {
            getEditor(e).getCompleter().pageDown();
        }
    }

    private class ActionTab extends OverrideAction {

        protected void overrideEvent(ActionEvent e) {
            Database db = getEditor(e).getDatabase();
            CompleterItem[] items = _completer.getSelectedItems();
            try {
                if (items != null && items.length == 1) {
                    Collection<CompleterItem> completerItems = db.getResolver().resolveChildren(db, items[0].getObject());
                    if (completerItems.size() > 0) {
                        _completer.setItems(completerItems);
                    }
                }
            } catch (Exception e1) {
                Utils.debug("TokenEditor$ActionTab.overrideEvent: " + e1);
            }
        }
    }

    private class MyCaretListener implements CaretListener {
        protected int _lastCaretPosition = -1;

        public void caretUpdate(CaretEvent e) {
            if (_lastCaretPosition != -1 && Math.abs(e.getDot() - _lastCaretPosition) != 1) {
                _completer.closeCompleter();
            }
            _lastCaretPosition = e.getDot();
        }
    }

    private class MyDocumentListener implements DocumentListener {
        private boolean _dotDeleted = false;

        public void insertUpdate(DocumentEvent e) {
            updateCompleter(e);
        }

        private void updateCompleter(DocumentEvent e) {
            try {
                Document doc = getDocument();
                String text = doc.getText(0, doc.getLength());
                EventType type = e.getType();
                Database db = getDatabase();
                if ( null != db ) {
                    int delta = EventType.REMOVE == type ? 1 : 0;
                    int offset = e.getOffset() - delta;
                    if (_completer.isVisible()) {
                        CharSequence filter = WordScanner.getCompletingWord(text, offset);
                        if (!_dotDeleted && null != filter) {
                            _completer.setFilter(filter);
                        } else {
                            try {
                                Collection<CompleterItem> items = db.getResolver().resolve(db, text, offset);
                                _completer.setItems(items);
                                _completer.setFilter(filter);
                            } catch (Exception e1) {
                                Utils.debug("TokenEditor$MyDocumentListener.updateCompleter: " + e1);
                            }
                        }
                    } else if (type == EventType.INSERT) {
                        try {
                            CharSequence filter = WordScanner.getCompletingWord(text, offset);
                            if (filter != null && filter.charAt(filter.length() - 1) == '.') {
                                try {
                                    Collection<CompleterItem> items = db.getResolver().resolve(db, text, offset);
                                    _completer.setItems(items);
                                    _completer.showCompleter();
                                } catch (Exception e1) {
                                    Utils.debug("TokenEditor$MyDocumentListener.updateCompleter: " + e1);
                                }
                            }
                        } catch (IllegalStateException e1) {
                            // ignore to no suggestion exception
                        }
                    }
                }
            } catch (BadLocationException e1) {
                Utils.debug("TokenEditor$MyDocumentListener.updateCompleter: " + e1);
            }
        }

        public void removeUpdate(DocumentEvent e) {
            updateCompleter(e);
        }

        public void changedUpdate(DocumentEvent e) {
        }

        public void setDeletingDot(boolean b) {
            _dotDeleted = b;
        }
    }

    private static class ActionUndo extends AbstractAction {
        public void actionPerformed(ActionEvent e) {
            TokenEditor editor = (TokenEditor) e.getSource();
            editor.undo();
        }
    }
}
